from typing import Optional

from .raster_interface import RasterInterface

try:
    import rasterio
except ImportError:
    rasterio = None


class RasterIORasterInterface(RasterInterface):
    def __init__(self, path):
        if rasterio is None:
            raise ImportError("This raster check requires rasterio")
        super().__init__(path)

    @staticmethod
    def available():
        return rasterio is not None

    def _open(self):
        with rasterio.Env(
            CPL_VSIL_CURL_USE_HEAD="NO",
            GDAL_DISABLE_READDIR_ON_OPEN="YES",
        ):
            try:
                self._dataset = rasterio.open(self.path, "r")
            except rasterio.RasterioIOError:
                self._dataset = None

    def _close(self):
        if self._dataset is not None:
            self._dataset.close()
            self._dataset = None

    @property
    def is_valid_geotiff(self):
        return self._dataset is not None and self._dataset.driver == "GTiff"

    @property
    def band_count(self):
        return self._dataset.count

    @property
    def has_projection(self) -> bool:
        return self._dataset.crs is not None

    @property
    def is_geographic(self) -> bool:
        return self._dataset.crs.is_geographic

    @property
    def epsg_code(self) -> Optional[int]:
        return self._dataset.crs.to_epsg()

    @property
    def pixel_size(self):
        gt = self._dataset.get_transform()
        if gt is None:
            return None, None
        else:
            return abs(gt[1]), abs(gt[5])

    @property
    def min_max(self):
        if self.band_count == 0:
            return None, None
        try:
            statistics = self._dataset.statistics(1, approx=False, clear_cache=True)
        except Exception as e:
            if "no valid pixels found" in str(e).lower():
                raise self.NoData()
            else:
                raise e
        return statistics.min, statistics.max

    @property
    def shape(self):
        return (self._dataset.height, self._dataset.width)

    @property
    def compression(self) -> str:
        return self._dataset.compression.value



from unittest.mock import MagicMock, patch

import pytest
from geoalchemy2.elements import WKBElement
from threedi_schema.domain.constants import InflowType

from threedi_modelchecker.checks.base import CheckLevel
from threedi_modelchecker.exporters import (
    export_with_geom,
    generate_csv_table,
    generate_rst_table,
)


@pytest.fixture
def fake_check_warning():
    fake_check = MagicMock()
    fake_check.level = CheckLevel.WARNING
    fake_check.error_code = 2
    fake_check.description.return_value = (
        "This sample message has code 2 and level WARNING"
    )
    return fake_check


@pytest.fixture
def fake_check_error():
    fake_check = MagicMock()
    fake_check.level = CheckLevel.ERROR
    fake_check.error_code = 1234
    fake_check.description.return_value = (
        "This sample message has code 1234 and level ERROR"
    )
    return fake_check


@pytest.fixture
def fake_check_info():
    fake_check = MagicMock()
    fake_check.level = CheckLevel.INFO
    fake_check.error_code = 12
    fake_check.description.return_value = (
        "This sample message has code 12 and level INFO"
    )
    return fake_check


@pytest.fixture
def fake_checks(fake_check_warning, fake_check_error, fake_check_info):
    return [fake_check_warning, fake_check_error, fake_check_info]


def test_generate_rst_table(fake_checks):
    correct_rst_result = (
        ".. list-table:: Executed checks\n"
        + "   :widths: 10 20 40\n   :header-rows: 1\n\n"
        + "   * - Check number\n"
        + "     - Check level\n"
        + "     - Check message\n"
        + "   * - 0002\n"
        + "     - Warning\n"
        + "     - This sample message has code 2 and level WARNING\n"
        + "   * - 0012\n"
        + "     - Info\n"
        + "     - This sample message has code 12 and level INFO\n"
        + "   * - 1234\n"
        + "     - Error\n"
        + "     - This sample message has code 1234 and level ERROR"
    )
    rst_result = generate_rst_table(fake_checks)
    assert rst_result == correct_rst_result


def test_generate_csv_table(fake_checks):
    correct_csv_result = (
        '"error_code","level","description"\r\n'
        + '2,"WARNING","This sample message has code 2 and level WARNING"\r\n'
        + '12,"INFO","This sample message has code 12 and level INFO"\r\n'
        + '1234,"ERROR","This sample message has code 1234 and level ERROR"\r\n'
    )
    csv_result = generate_csv_table(fake_checks)
    assert csv_result == correct_csv_result


def test_export_with_geom(fake_check_error):
    fake_check_error.column.name = "foo"

    # First test case - no geom
    error_row_no_geom = MagicMock()
    error_row_no_geom.id = 1337
    error_row_no_geom.foo = "bar"
    del error_row_no_geom.geom

    # Second test case - with geom
    error_row_geom = MagicMock()
    error_row_geom.geom = MagicMock(spec=WKBElement)

    # Third test case - WKBElement in value attribute
    error_row_geom_value = MagicMock()
    error_row_geom_value.geom = MagicMock(spec=WKBElement)

    # Mock class so isinstance check works
    class CustomWKBElement:
        __class__ = WKBElement

    error_row_geom_value.foo = CustomWKBElement()

    # Fourth test case value is an enum
    error_row_enum = MagicMock()
    error_row_enum.foo = InflowType.NO_INFLOW

    # Setup to_shape patch before calling export_with_geom
    with patch("threedi_modelchecker.exporters.to_shape") as mock_to_shape:
        # Configure mock to return a shape with 'wkt' property
        mock_shape = MagicMock()
        mock_shape.wkt = "wkt"
        mock_to_shape.return_value = mock_shape
        result = export_with_geom(
            [
                (fake_check_error, error_row_no_geom),
                (fake_check_error, error_row_geom),
                (fake_check_error, error_row_geom_value),
                (fake_check_error, error_row_enum),
            ]
        )

    # Assertions for first and second row remain the same
    assert result[0].name == fake_check_error.level.name
    assert result[0].code == fake_check_error.error_code
    assert result[0].id == 1337
    assert result[0].table == fake_check_error.table.name
    assert result[0].description == fake_check_error.description()
    assert result[0].column == "foo"
    assert result[0].value == "bar"
    assert result[0].geom is None
    assert result[1].geom == error_row_geom.geom
    assert result[2].value == "wkt"
    assert result[3].value == "No inflow"



import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Literal, NamedTuple

import pyproj
from geoalchemy2.functions import ST_Distance, ST_Length
from sqlalchemy import (
    and_,
    case,
    cast,
    distinct,
    func,
    not_,
    REAL,
    select,
    text,
    union_all,
)
from sqlalchemy.orm import aliased, Query, Session
from threedi_schema.domain import constants, models

from .base import BaseCheck, CheckLevel
from .cross_section_definitions import cross_section_configuration_for_record


class CorrectAggregationSettingsExist(BaseCheck):
    """Check if aggregation settings are correctly filled with aggregation_method and flow_variable as required"""

    def __init__(
        self,
        aggregation_method: constants.AggregationMethod,
        flow_variable: constants.FlowVariable,
        *args,
        **kwargs,
    ):
        super().__init__(column=models.ModelSettings.id, *args, **kwargs)
        self.aggregation_method = aggregation_method.value
        self.flow_variable = flow_variable.value

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        global_settings = self.to_check(session)
        correctly_defined = session.execute(
            select(models.AggregationSettings).filter(
                models.AggregationSettings.aggregation_method
                == self.aggregation_method,
                models.AggregationSettings.flow_variable == self.flow_variable,
            )
        ).all()
        return global_settings.all() if len(correctly_defined) == 0 else []

    def description(self) -> str:
        return (
            "To use the water balance tool, aggregation_settings should have a row where "
            f"aggregation_method is {self.aggregation_method} and flow_variable is {self.flow_variable}."
        )


class CrossSectionSameConfigurationCheck(BaseCheck):
    """Check the cross-sections on the object are either all open or all closed."""

    def first_number_in_spaced_string(self, spaced_string):
        """return the first number in a space-separated string like '1 2 3'"""
        return cast(
            func.substr(
                spaced_string,
                1,
                func.instr(spaced_string, " ") - 1,
            ),
            REAL,
        )

    def last_number_in_spaced_string(self, spaced_string):
        """return the last number in a space-separated string like '1 2 3'"""
        return cast(
            func.replace(
                spaced_string,
                func.rtrim(
                    spaced_string,
                    func.replace(spaced_string, " ", ""),
                ),
                "",
            ),
            REAL,
        )

    def get_first_in_str(self, col, sep):
        return func.substr(
            col,
            1,
            func.instr(col, sep) - 1,
        )

    def get_last_in_str(self, col, sep):
        return func.replace(
            col,
            func.rtrim(
                col,
                func.replace(col, sep, ""),
            ),
            "",
        )

    def first_row_width(self):
        first_row = self.get_first_in_str(
            models.CrossSectionLocation.cross_section_table, "\n"
        )
        return cast(self.get_first_in_str(first_row, ","), REAL)

    def first_row_height(self):
        first_row = self.get_first_in_str(
            models.CrossSectionLocation.cross_section_table, "\n"
        )
        return cast(self.get_last_in_str(first_row, ","), REAL)

    def last_row_width(self):
        last_row = self.get_last_in_str(
            models.CrossSectionLocation.cross_section_table, "\n"
        )
        return cast(self.get_first_in_str(last_row, ","), REAL)

    def last_row_height(self):
        last_row = self.get_last_in_str(
            models.CrossSectionLocation.cross_section_table, "\n"
        )
        return cast(self.get_last_in_str(last_row, ","), REAL)

    def configuration_type(
        self, shape, first_width, last_width, first_height, last_height
    ):
        return case(
            (
                (
                    (shape.in_([0, 2, 3, 8]))
                    | (shape.in_([5, 6]) & (last_width == 0))
                    | (
                        (shape == 7)
                        & (first_width == last_width)
                        & (first_height == last_height)
                    )
                ),
                "closed",
            ),
            (
                (
                    (shape == 1)
                    | ((shape.in_([5, 6]) & (last_width > 0)))
                    | (
                        (shape == 7)
                        & ((first_width != last_width) | (first_height != last_height))
                    )
                ),
                "open",
            ),
            else_="open",
        )

    def get_invalid(self, session):
        # find all tabulated cross sections
        cross_sections_tab = select(
            models.CrossSectionLocation.id.label("cross_section_id"),
            models.CrossSectionLocation.channel_id,
            models.CrossSectionLocation.cross_section_shape,
            models.CrossSectionLocation.cross_section_table,
            self.first_row_width().label("first_width"),
            self.first_row_height().label("first_height"),
            self.last_row_width().label("last_width"),
            self.last_row_height().label("last_height"),
        ).where(models.CrossSectionLocation.cross_section_shape.in_([5, 6, 7]))
        # find all non tabulated cross sections
        cross_sections_notab = select(
            models.CrossSectionLocation.id.label("cross_section_id"),
            models.CrossSectionLocation.channel_id,
            models.CrossSectionLocation.cross_section_shape,
            models.CrossSectionLocation.cross_section_table,
            models.CrossSectionLocation.cross_section_width.label("first_width"),
            models.CrossSectionLocation.cross_section_height.label("first_height"),
            models.CrossSectionLocation.cross_section_width.label("last_width"),
            models.CrossSectionLocation.cross_section_height.label("last_height"),
        ).where(~models.CrossSectionLocation.cross_section_shape.in_([5, 6, 7]))
        # combine the above two queries to get all cross sections
        cross_sections = union_all(cross_sections_tab, cross_sections_notab).subquery()
        cross_sections_with_configuration = select(
            cross_sections.c.cross_section_id,
            cross_sections.c.cross_section_shape,
            cross_sections.c.last_width,
            cross_sections.c.channel_id,
            self.configuration_type(
                shape=cross_sections.c.cross_section_shape,
                first_width=cross_sections.c.first_width,
                last_width=cross_sections.c.last_width,
                first_height=cross_sections.c.first_height,
                last_height=cross_sections.c.last_height,
            ).label("configuration"),
        ).subquery()

        filtered_cross_sections = (
            select(cross_sections_with_configuration)
            .group_by(cross_sections_with_configuration.c.channel_id)
            .having(
                func.count(distinct(cross_sections_with_configuration.c.configuration))
                > 1
            )
            .subquery()
        )
        return (
            self.to_check(session)
            .filter(self.column == filtered_cross_sections.c.channel_id)
            .all()
        )

    def description(self):
        return f"{self.column_name} has both open and closed cross-sections along its length. All cross-sections on a {self.column_name} object must be either open or closed."


class Use0DFlowCheck(BaseCheck):
    """Check that when use_0d_flow in global settings is configured to 1 or to
    2, there is at least one impervious surface or surfaces respectively.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            column=models.SimulationTemplateSettings.use_0d_inflow, *args, **kwargs
        )

    def get_invalid(self, session):
        settings = session.query(models.SimulationTemplateSettings).one_or_none()
        if settings is None:
            return []
        use_0d_flow = settings.use_0d_inflow
        if use_0d_flow != constants.InflowType.NO_INFLOW:
            surface_count = session.query(func.count(models.Surface.id)).scalar()
            if surface_count == 0:
                return [settings]
        return []

    def description(self):
        return (
            f"When {self.column_name} is used, there should exist at least one surface."
        )


class ConnectionNodes(BaseCheck):
    """Check that all connection nodes are connected to at least one of the
    following objects:
    - Culvert
    - Channel
    - Pipe
    - Orifice
    - Pumpstation
    - Weir
    """

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ConnectionNode.id, *args, **kwargs)

    def get_invalid(self, session):
        raise NotImplementedError


class ConnectionNodesLength(BaseCheck):
    """Check that the distance between `start_node` and `end_node` is at least
    `min_distance`.
    """

    def __init__(
        self,
        start_node,
        end_node,
        min_distance: float,
        recommended_distance: float = 1.0,
        *args,
        **kwargs,
    ):
        """

        :param start_node: column name of the start node
        :param end_node: column name of the end node
        :param min_distance: minimum required distance between start and end node
        """
        super().__init__(*args, **kwargs)
        self.start_node = start_node
        self.end_node = end_node
        self.min_distance = min_distance
        self.recommended_distance = recommended_distance

    def get_invalid(self, session):
        start_node = aliased(models.ConnectionNode)
        end_node = aliased(models.ConnectionNode)
        q = (
            self.to_check(session)
            .join(start_node, start_node.id == self.start_node)
            .join(end_node, end_node.id == self.end_node)
            .filter(ST_Distance(start_node.geom, end_node.geom) < self.min_distance)
        )
        return list(q.with_session(session).all())

    def description(self) -> str:
        return (
            f"The length of {self.table} is "
            f"very short (< {self.min_distance}). A length of at least {self.recommended_distance} m is recommended to avoid timestep reduction."
        )


class ConnectionNodesDistance(BaseCheck):
    """Check that the distance between connection nodes is above a certain
    threshold
    """

    def __init__(
        self, minimum_distance: float, level=CheckLevel.WARNING, *args, **kwargs
    ):
        """
        :param minimum_distance: threshold distance in degrees
        """
        super().__init__(column=models.ConnectionNode.id, level=level, *args, **kwargs)
        self.minimum_distance = minimum_distance

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        """
        The query makes use of the SpatialIndex so we won't have to calculate the
        distance between all connection nodes.
        """
        query = text(
            """
            SELECT *
            FROM connection_node AS cn1, connection_node AS cn2
            WHERE ST_Distance(cn1.geom, cn2.geom) < :minimum_distance
            AND cn1.ROWID != cn2.ROWID
            AND cn2.ROWID IN (
                SELECT ROWID
                FROM rtree_connection_node_geom
                WHERE (
                    maxx >= ST_MinX(ST_Buffer(cn1.geom, :buffer_distance))
                    AND minx <= ST_MaxX(ST_Buffer(cn1.geom, :buffer_distance))
                    AND maxy >= ST_MinY(ST_Buffer(cn1.geom, :buffer_distance))
                    AND miny <= ST_MaxY(ST_Buffer(cn1.geom, :buffer_distance))
                )
            )
            """
        ).bindparams(
            minimum_distance=self.minimum_distance,
            buffer_distance=self.minimum_distance / 2,
        )
        # get ids for invalid rows
        result_ids = [row.id for row in session.execute(query).fetchall()]
        # use sqlalchemy to return invalid rows with proper WKBElement for geometries
        return (
            session.query(models.ConnectionNode)
            .filter(models.ConnectionNode.id.in_(result_ids))
            .all()
        )

    def description(self) -> str:

        return (
            f"The connection_node is within {self.minimum_distance * 100} cm of "
            f"another connection_node."
        )


class ChannelManholeLevelCheck(BaseCheck):
    """Check that the reference_level of a channel is higher than or equal to the bottom_level of a manhole
    connected to the channel as measured at the cross-section closest to the manhole. This check runs if the
    manhole is on the channel's starting node.
    """

    def __init__(
        self,
        level: CheckLevel = CheckLevel.INFO,
        nodes_to_check: Literal["start", "end"] = "start",
        *args,
        **kwargs,
    ):
        """
        :param level: severity of the check, defaults to CheckLevel.INFO. Options are
        in checks.base.CheckLevel
        :param nodes_to_check: whether to check for manholes at the start of the channel
        or at the end of the channel. Options are "start" and "end", defaults to "start"
        """
        if nodes_to_check not in ["start", "end"]:
            raise ValueError("nodes_to_check must be 'start' or 'end'")
        super().__init__(column=models.Channel.id, level=level, *args, **kwargs)
        self.nodes_to_check = nodes_to_check

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        """
        This query does the following:
        channel_with_cs_locations       : left join between cross_sections and channels, to get a table containing
                                          all cross-sections and the channels they lie on
        channels_with_manholes          : join between channel_with_cs_locations and manholes, to get all channels with
                                          a manhole on the channel's start node if self.nodes_to_check == "start", or all
                                          channels with a manhole on the channel's end node if self.nodes_to_check == "start".
        channels_manholes_level_checked : filter the query on invalid entries; that is, entries where the cross-section
                                          reference level is indeed lower than the manhole bottom level. having is used instead
                                          of filter because the query being filtered is a aggregate produced by groupby.
        """
        if self.nodes_to_check == "start":
            func_agg = func.MIN
            connection_node_id_col = models.Channel.connection_node_id_start
        else:
            func_agg = func.MAX
            connection_node_id_col = models.Channel.connection_node_id_end

        channels_with_cs_locations = (
            session.query(
                models.Channel.id,
                models.Channel.geom,
                models.CrossSectionLocation,
                func_agg(
                    func.Line_Locate_Point(
                        models.Channel.geom, models.CrossSectionLocation.geom
                    )
                ),
            )
            .join(
                models.CrossSectionLocation,
                models.CrossSectionLocation.channel_id == models.Channel.id,
            )
            .group_by(models.Channel.id)
        )

        channels_with_manholes = channels_with_cs_locations.join(
            models.ConnectionNode, models.ConnectionNode.id == connection_node_id_col
        )

        channels_manholes_level_checked = channels_with_manholes.having(
            models.CrossSectionLocation.reference_level
            < models.ConnectionNode.bottom_level
        )

        return channels_manholes_level_checked.all()

    def description(self) -> str:
        return (
            f"The connection_node.bottom_level at the {self.nodes_to_check} of this channel is higher than the "
            "cross_section_location.reference_level closest to the connection-node. This will be "
            "automatically fixed in threedigrid-builder."
        )


class OpenChannelsWithNestedNewton(BaseCheck):
    """Checks whether the model has any closed cross-section in use when the
    NumericalSettings.use_nested_newton is turned off.

    See https://github.com/nens/threeditoolbox/issues/522
    """

    def __init__(self, column, level=CheckLevel.WARNING, *args, **kwargs):
        super().__init__(
            # column=table.id,
            column=column,
            level=level,
            filters=Query(models.NumericalSettings)
            .filter(models.NumericalSettings.use_nested_newton == 0)
            .exists(),
            *args,
            **kwargs,
        )
        # self.table = table

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        invalids = []
        for record in self.to_check(session):
            _, _, configuration = cross_section_configuration_for_record(record)

            if configuration == "closed":
                invalids.append(record)
        return invalids

    def description(self) -> str:
        return (
            f"{self.column_name} has a closed cross section definition while "
            f"NumericalSettings.use_nested_newton is switched off. "
            f"This gives convergence issues. We recommend setting use_nested_newton = 1."
        )


class BoundaryCondition1DObjectNumberCheck(BaseCheck):
    """Check that the number of connected objects to 1D boundary connections is 1."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            column=models.BoundaryCondition1D.connection_node_id, *args, **kwargs
        )

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        invalid_ids = []
        for bc in self.to_check(session).all():
            total_objects = 0
            for table in [
                models.Channel,
                models.Pipe,
                models.Culvert,
                models.Orifice,
                models.Weir,
            ]:
                total_objects += (
                    session.query(table)
                    .filter(table.connection_node_id_start == bc.connection_node_id)
                    .count()
                )
                total_objects += (
                    session.query(table)
                    .filter(table.connection_node_id_end == bc.connection_node_id)
                    .count()
                )
            if total_objects != 1:
                invalid_ids.append(bc.id)

        return (
            self.to_check(session)
            .filter(models.BoundaryCondition1D.id.in_(invalid_ids))
            .all()
        )

    def description(self) -> str:
        return "1D boundary condition should be connected to exactly one object."


@dataclass
class IndexMissingRecord:
    id: int
    table_name: str
    column_name: str


class SpatialIndexCheck(BaseCheck):
    """Checks whether a spatial index is present and valid"""

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        result = session.execute(
            text(
                f"""
            SELECT EXISTS(
                SELECT 1 FROM sqlite_master
                WHERE type='table'
                AND name='rtree_{self.column.table.name}_{self.column.name}'
            );
        """
            )
        ).scalar()
        if result == 1:
            return []
        else:
            return [
                IndexMissingRecord(
                    id=1,
                    table_name=self.column.table.name,
                    column_name=self.column.name,
                )
            ]

    def description(self) -> str:
        return f"{self.column_name} has no valid spatial index, which is required for some checks"


class PotentialBreachStartEndCheck(BaseCheck):
    """Check that a potential breach is exactly on or >=1 m from a linestring start/end."""

    def __init__(self, *args, **kwargs):
        self.min_distance = kwargs.pop("min_distance")

        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        linestring = models.Channel.geom
        tol = self.min_distance
        breach_point = func.Line_Locate_Point(
            linestring, func.ST_PointN(self.column, 1)
        )
        dist_1 = breach_point * ST_Length(linestring)
        dist_2 = (1 - breach_point) * ST_Length(linestring)
        return (
            self.to_check(session)
            .join(models.Channel, self.table.c.channel_id == models.Channel.id)
            .filter(((dist_1 > 0) & (dist_1 < tol)) | ((dist_2 > 0) & (dist_2 < tol)))
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} must be exactly on or >= {self.min_distance} m from a start or end channel vertex"


class PotentialBreachInterdistanceCheck(BaseCheck):
    """Check that a potential breaches are exactly on the same place or >=1 m apart."""

    def __init__(self, *args, **kwargs):
        self.min_distance = kwargs.pop("min_distance")
        assert "filters" not in kwargs

        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        # this query is hard to get performant; we do a hybrid sql / Python approach

        # First fetch the position of each potential breach per channel
        def get_position(point, linestring):
            breach_point = func.Line_Locate_Point(linestring, func.ST_PointN(point, 1))
            return (breach_point * ST_Length(linestring)).label("position")

        potential_breaches = sorted(
            session.query(self.table, get_position(self.column, models.Channel.geom))
            .join(models.Channel, self.table.c.channel_id == models.Channel.id)
            .all(),
            key=lambda x: (x.channel_id, x[-1]),
        )

        invalid = []
        prev_channel_id = -9999
        prev_position = -1.0
        for breach in potential_breaches:
            if breach.channel_id != prev_channel_id:
                prev_channel_id, prev_position = breach.channel_id, breach.position
                continue
            if breach.position == prev_position:
                continue
            if (breach.position - prev_position) <= self.min_distance:
                invalid.append(breach)
        return invalid

    def description(self) -> str:
        return f"{self.column_name} must be more than {self.min_distance} m apart (or exactly on the same position)"


class PumpStorageTimestepCheck(BaseCheck):
    """Check that a pumpstation will not empty its storage area within one timestep"""

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        return (
            session.query(models.Pump)
            .join(
                models.ConnectionNode,
                models.Pump.connection_node_id == models.ConnectionNode.id,
            )
            .filter(
                (models.ConnectionNode.storage_area != None)
                & (
                    (
                        # calculate how many seconds the pumpstation takes to empty its storage: (storage * height)/pump capacity
                        (
                            # Arithmetic operations on None return None, so without this
                            # conditional type cast, no invalid results would be returned
                            # even if the storage_area was set to None.
                            models.ConnectionNode.storage_area
                            * (models.Pump.start_level - models.Pump.lower_stop_level)
                        )
                    )
                    / (models.Pump.capacity / 1000)
                    < Query(models.TimeStepSettings.time_step).scalar_subquery()
                )
            )
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} will empty its storage faster than one timestep, which can cause simulation instabilities"


class SurfaceNodeInflowAreaCheck(BaseCheck):
    """Check that total inflow area per connection node is no larger than 10000 square metres"""

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ConnectionNode.id, *args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        surfaces = (
            select(models.SurfaceMap.connection_node_id)
            .select_from(models.SurfaceMap)
            .join(
                models.Surface,
                models.SurfaceMap.surface_id == models.Surface.id,
            )
            .group_by(models.SurfaceMap.connection_node_id)
            .having(func.sum(models.Surface.area) > 10000)
        ).subquery()

        return (
            session.query(models.ConnectionNode)
            .filter(models.ConnectionNode.id == surfaces.c.connection_node_id)
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} has a an associated inflow area larger than 10000 m2; this might be an error."


class PerviousNodeInflowAreaCheck(BaseCheck):
    """Check that total inflow area per connection node is no larger than 10000 square metres"""

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ConnectionNode.id, *args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        pervious_surfaces = (
            select(models.SurfaceMap.connection_node_id)
            .select_from(models.SurfaceMap)
            .join(
                models.Surface,
                models.SurfaceMap.surface_id == models.Surface.id,
            )
            .group_by(models.SurfaceMap.connection_node_id)
            .having(func.sum(models.Surface.area) > 10000)
        ).subquery()

        return (
            session.query(models.ConnectionNode)
            .filter(models.ConnectionNode.id == pervious_surfaces.c.connection_node_id)
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} has a an associated inflow area larger than 10000 m2; this might be an error."


class InflowNoFeaturesCheck(BaseCheck):
    """Check that the surface table in the global use_0d_inflow setting contains at least 1 feature."""

    def __init__(self, *args, feature_table, condition=True, **kwargs):
        super().__init__(*args, column=models.ModelSettings.id, **kwargs)
        self.feature_table = feature_table
        self.condition = condition

    def get_invalid(self, session: Session):
        surface_table_length = session.execute(
            select(func.count(self.feature_table.id))
        ).scalar()
        return (
            session.query(models.ModelSettings)
            .filter(self.condition, surface_table_length == 0)
            .all()
        )

    def description(self) -> str:
        return f"model_settings.use_0d_inflow is set to use {self.feature_table.__tablename__}, but {self.feature_table.__tablename__} does not contain any features."


class NodeSurfaceConnectionsCheck(BaseCheck):
    """Check that no more than 50 surfaces are mapped to a connection node"""

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ConnectionNode.id, *args, **kwargs)
        self.surface_column = models.SurfaceMap

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        if self.surface_column is None:
            return []
        overloaded_connections = (
            select(models.SurfaceMap.connection_node_id)
            .group_by(models.SurfaceMap.connection_node_id)
            .having(func.count(models.SurfaceMap.connection_node_id) > 50)
        )

        return (
            self.to_check(session)
            .filter(models.ConnectionNode.id.in_(overloaded_connections))
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} has more than 50 surface areas mapped to it; this might be an error."


class FeatureClosedCrossSectionCheck(BaseCheck):
    """
    Check if feature has a closed cross-section
    """

    def get_invalid(self, session):
        invalids = []
        for record in self.to_check(session):
            _, _, configuration = cross_section_configuration_for_record(record)

            # Pipes and culverts should generally have a closed cross-section
            if configuration == "open":
                invalids.append(record)

        return invalids

    def description(self):
        return f"{self.column_name} has an open cross-section, which is unusual for this feature. Please make sure this is not a mistake."


class DefinedAreaCheck(BaseCheck):
    """Check if the value in the 'area' column matches the surface area of 'geom'"""

    def __init__(self, *args, max_difference=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_difference = max_difference

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        all_results = select(
            self.table.c.id,
            self.table.c.area,
            self.table.c.geom,
            func.ST_Area(self.table.c.geom).label("calculated_area"),
        ).subquery()
        return (
            session.query(all_results)
            .filter(
                func.abs(all_results.c.area - all_results.c.calculated_area)
                > self.max_difference
            )
            .all()
        )

    def description(self):
        return f"{self.column_name} has a {self.column_name} (used in the simulation) differing from its geometrical area by more than 1 m2"


class BetaColumnsCheck(BaseCheck):
    """Check that no beta columns were used in the database"""

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        return session.query(self.table).filter(self.column.isnot(None)).all()

    def description(self) -> str:
        return f"{self.column_name} is a beta feature, which is still under development; please do not use it yet."


class BetaValuesCheck(BaseCheck):
    """Check that no beta features were used in the database"""

    def __init__(
        self,
        column,
        values: list = [],
        filters=None,
        level=CheckLevel.ERROR,
        error_code=0,
    ):
        super().__init__(column, filters, level, error_code)
        self.values = values

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        return session.query(self.table).filter(self.column.in_(self.values)).all()

    def description(self) -> str:
        return f"The value you have used for {self.column_name} is still in beta; please do not use it yet."


class AllPresentVegetationParameters(BaseCheck):
    """Check if all or none vegetation values are defined in the CrossSectionLocation table"""

    def __init__(self, *args, **kwargs):
        self.columns = [
            models.CrossSectionLocation.vegetation_drag_coefficient,
            models.CrossSectionLocation.vegetation_height,
            models.CrossSectionLocation.vegetation_stem_diameter,
            models.CrossSectionLocation.vegetation_stem_density,
        ]
        super().__init__(*args, **kwargs)

    def get_invalid(self, session):
        # Create filters that find all rows where all or none of the values are present
        filter_condition_all = and_(
            *[(col != None) & (col != "") for col in self.columns]
        )
        filter_condition_none = and_(
            *[(col == None) | (col == "") for col in self.columns]
        )
        # Return all rows where neither all or none values are present
        records = (
            session.query(models.CrossSectionLocation)
            .filter(
                models.CrossSectionLocation.friction_type.is_(
                    constants.FrictionType.CHEZY
                )
            )
            .filter(
                models.CrossSectionLocation.cross_section_shape.is_(
                    constants.CrossSectionShape.TABULATED_YZ
                )
            )
        )

        return (
            records.filter(not_(filter_condition_all))
            .filter(not_(filter_condition_none))
            .all()
        )


class SettingsPresentCheck(BaseCheck, ABC):
    def __init__(
        self,
        column,
        settings_tables,
        filters=None,
        level=CheckLevel.ERROR,
        error_code=0,
    ):
        super().__init__(column, filters, level, error_code)
        self.settings_tables = settings_tables

    @abstractmethod
    def get_all_results(self, session):
        pass

    @abstractmethod
    def get_table_condition(self, session):
        pass

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        # more than 1 row should be caught by another check
        all_results = self.get_all_results(session)
        not_use_cols = len(all_results) == 1
        # return as invalid if the use_col is true but none of the associated tables are actually used
        if not_use_cols and self.get_table_condition(session):
            return all_results
        return []


class UnusedSettingsPresentCheck(SettingsPresentCheck):

    def get_all_results(self, session):
        return self.to_check(session).filter(self.column == False).all()

    def get_table_condition(self, session):
        return any(session.query(table).count() > 0 for table in self.settings_tables)

    def description(self) -> str:
        msg = f"{self.column_name} in {self.table.name} is not set to True but "
        if len(self.settings_tables) == 1:
            msg += "{self.settings_tables[0].__tablename__} is not empty"
        else:
            msg += (
                "["
                + ",".join(table.__tablename__ for table in self.settings_tables)
                + "] are not empty"
            )
        return msg


class UsedSettingsPresentCheck(SettingsPresentCheck):

    def get_all_results(self, session):
        return self.to_check(session).filter(self.column == True).all()

    def get_table_condition(self, session):
        return all(session.query(table).count() == 0 for table in self.settings_tables)

    def description(self) -> str:
        msg = f"{self.column_name} in {self.table.name} is set to True but "
        if len(self.settings_tables) == 1:
            msg += "{self.settings_tables[0].__tablename__} is empty"
        else:
            msg += (
                "["
                + ",".join(table.__tablename__ for table in self.settings_tables)
                + "] are empty"
            )
        return msg


class UsedSettingsPresentCheckSingleTable(UsedSettingsPresentCheck):
    def __init__(
        self,
        column,
        settings_table,
        filters=None,
        level=CheckLevel.ERROR,
        error_code=0,
    ):
        super().__init__(column, [settings_table], filters, level, error_code)


class MaxOneRecordCheck(BaseCheck):
    def __init__(self, column, filters=None, level=CheckLevel.ERROR, error_code=0):
        super().__init__(column, filters, level, error_code)
        self.observed_length = 0

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        # return mock list in case the table is empty when it shouldn't be
        all_results = self.to_check(session).all()
        self.observed_length = len(all_results)
        if self.observed_length > 1:
            return all_results if self.observed_length > 0 else ["foo"]
        else:
            return []

    def description(self) -> str:
        return (
            f"{self.table.name} has {self.observed_length} rows, "
            f"but should have at most 1 row."
        )


class TagsValidCheck(BaseCheck):
    def get_invalid(self, session):
        invalids = []
        try:
            for record in self.to_check(session).filter(
                (self.column != None) & (self.column != "")
            ):
                query = f"SELECT id FROM tags WHERE id IN ({getattr(record, self.column.name)})"
                match_rows = session.connection().execute(text(query)).fetchall()
                found_idx = {row[0] for row in match_rows}
                req_idx = {int(x) for x in getattr(record, self.column.name).split(",")}
                if found_idx != req_idx:
                    invalids.append(record)
        except (
            ValueError
        ):  # in case of badly formatted tags; handled by ListOfIntsCheck
            pass
        return invalids

    def description(self) -> str:
        return f"{self.table.name}.{self.column} refers to tag ids that are not present in Tags, "


class TableStrCheck(BaseCheck):
    def __init__(
        self, column, pattern, filters=None, level=CheckLevel.ERROR, error_code=0
    ):
        self.pattern = pattern
        super().__init__(
            column=column, filters=filters, level=level, error_code=error_code
        )

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        # return mock list in case the table is empty when it shouldn't be
        invalids = []
        for record in self.to_check(session).all():
            if re.match(self.pattern, getattr(record, self.column.name)) is None:
                invalids.append(record)
        return invalids


class TableControlActionTableCheckDefault(TableStrCheck):
    def __init__(self, level=CheckLevel.ERROR, error_code=0):
        # check for action_table for action_type != set_discharge_coefficients
        # expected format: multiple rows, separated by \n of "val,val"
        super().__init__(
            column=models.TableControl.action_table,
            pattern=r"^(-?\d+(\.\d+)?,-?\d+(\.\d+)?\n?)+$",
            filters=models.TableControl.action_type
            != constants.TableControlActionTypes.set_discharge_coefficients,
            level=level,
            error_code=error_code,
        )

        def description(self) -> str:
            return (
                f"{self.table.name}.{self.column} is not properly formatted."
                f"Expected one or more rows of: 'number, number number'"
            )


class TableControlActionTableCheckDischargeCoefficients(TableStrCheck):
    def __init__(self, level=CheckLevel.ERROR, error_code=0):
        # check for action_table for action_type = set_discharge_coefficients
        # expected format: multiple rows, separated by \n of "val,val val"
        super().__init__(
            column=models.TableControl.action_table,
            pattern=r"^(-?\d+(\.\d+)?,-?\d+(\.\d+)? -?\d+(\.\d+)?\n?)+$",
            filters=models.TableControl.action_type
            == constants.TableControlActionTypes.set_discharge_coefficients,
            level=level,
            error_code=error_code,
        )

    def description(self) -> str:
        return (
            f"{self.table.name}.{self.column} is not properly formatted."
            f"Expected one or more rows of: 'number, number'"
        )


class ControlHasSingleMeasureVariable(BaseCheck):
    def __init__(self, control_model, level=CheckLevel.ERROR, error_code=0):
        control_type_map = {
            models.TableControl: "table",
            models.MemoryControl: "memory",
        }
        self.control_type_name = control_type_map[control_model]
        super().__init__(
            column=control_model.id,
            level=level,
            error_code=error_code,
        )

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        invalid = []
        for record in self.to_check(session):
            res = (
                session.query(models.MeasureMap)
                .filter(
                    models.MeasureMap.control_type == self.control_type_name,
                    models.MeasureMap.control_id == record.id,
                )
                .join(
                    models.MeasureLocation,
                    models.MeasureMap.measure_location_id == models.MeasureLocation.id,
                )
                .with_entities(models.MeasureLocation.measure_variable)
            ).all()
            if len(res) == 0:
                continue
            first_measure_variable = res[0].measure_variable
            if not all(item[0] == first_measure_variable for item in res):
                invalid.append(record)
        return invalid

    def description(self) -> str:
        return f"{self.table.name} is mapped to measure locations with different measure variables"


class DWFDistributionBaseCheck(BaseCheck):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            column=models.DryWeatherFlowDistribution.distribution, *args, **kwargs
        )

    def get_distribution(self, record):
        str = getattr(record, self.column.name)
        return str.split(",")

    def get_distribution_values(self, record):
        return [float(x) for x in self.get_distribution(record)]


class DWFDistributionCSVFormatCheck(DWFDistributionBaseCheck):
    def get_invalid(self, session: Session) -> List[NamedTuple]:
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                self.get_distribution_values(record)
            except ValueError:
                invalids.append(record)
        return invalids

    def description(self) -> str:
        return f"{self.table.name}.{self.column_name} should contain a list of comma-separated numbers"


class DWFDistributionLengthCheck(DWFDistributionBaseCheck):
    def get_invalid(self, session: Session) -> List[NamedTuple]:
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            if len(self.get_distribution(record)) != 24:
                invalids.append(record)
        return invalids

    def description(self) -> str:
        return f"{self.table.name}.{self.column_name} should contain exactly 24 values"


class DWFDistributionSumCheck(DWFDistributionBaseCheck):
    def get_invalid(self, session: Session) -> List[NamedTuple]:
        invalids = []
        for record in self.to_check(session).filter(
            (self.column != None) & (self.column != "")
        ):
            try:
                values = self.get_distribution_values(record)
            except ValueError:
                # handled by DWFDistributionCSVFormatCheck
                continue
            if not (99.99 <= sum(values) <= 100.01):
                invalids.append(record)
        return invalids

    def description(self) -> str:
        return (
            f"The values in {self.table.name}.{self.column_name} should add up to 100 %"
        )


class ModelEPSGCheckValid(BaseCheck):
    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ModelSettings.id, *args, **kwargs)
        self.epsg_code = None

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        self.epsg_code = session.model_checker_context.epsg_ref_code
        if self.epsg_code is not None:
            try:
                pyproj.CRS.from_epsg(self.epsg_code)
            except pyproj.exceptions.CRSError:
                return self.to_check(session).all()
        return []

    def description(self) -> str:
        return f"Found invalid EPSG: {self.epsg_code}"


class ModelEPSGCheckProjected(BaseCheck):
    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ModelSettings.id, *args, **kwargs)
        self.epsg_code = None

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        self.epsg_code = session.model_checker_context.epsg_ref_code
        if self.epsg_code is not None:
            try:
                crs = pyproj.CRS.from_epsg(self.epsg_code)
            except pyproj.exceptions.CRSError:
                # handled by ModelEPSGCheckValid
                return []
            if not crs.is_projected:
                return self.to_check(session).all()
        return []

    def description(self) -> str:
        return f"EPSG {self.epsg_code} is not projected"


class ModelEPSGCheckUnits(BaseCheck):
    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ModelSettings.id, *args, **kwargs)
        self.epsg_code = None

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        self.epsg_code = session.model_checker_context.epsg_ref_code
        if self.epsg_code is not None:
            try:
                crs = pyproj.CRS.from_epsg(self.epsg_code)
            except pyproj.exceptions.CRSError:
                # handled by ModelEPSGCheckValid
                return []
            for ax in crs.axis_info:
                if not ax.unit_name == "metre":
                    return self.to_check(session).all()
        return []

    def description(self) -> str:
        return f"EPSG {self.epsg_code} is not fully defined in metres"


class GridRefinementPartialOverlap2DBoundaryCheck(BaseCheck):
    def get_invalid(self, session) -> List[NamedTuple]:
        invalid = []
        for record in self.to_check(session):
            # check how many grid refinement areas are crossed by the 2D boundary condition line
            overlapped_areas = len(
                session.execute(
                    select(models.GridRefinementArea).filter(
                        func.ST_Crosses(models.GridRefinementArea.geom, record.geom)
                    )
                ).all()
            )
            if overlapped_areas > 0:
                invalid.append(record)
        return invalid

    def description(self):
        return "2D boundary condition overlaps with grid refinement area. Make sure it is either completely contained by (within) or disjoint from the grid refinement polygon."



import pytest
from threedi_schema.domain import models

from threedi_modelchecker.checks.location import (
    ConnectionNodeLinestringLocationCheck,
    DWFMapLinestringLocationCheck,
    LinestringLocationCheck,
    MeasureMapLinestringMapLocationCheck,
    PointLocationCheck,
    PumpMapLinestringLocationCheck,
    SurfaceMapLinestringLocationCheck,
)
from threedi_modelchecker.tests import factories

SRID = 28992
POINT1 = "142742 473443"
POINT2 = "142743 473443"  # POINT1 + 1 meter
POINT3 = "142747 473443"  # POINT1 + 5 meter


@pytest.mark.parametrize(
    "bc_geom, nof_invalid", [(POINT1, 0), (POINT2, 0), (POINT3, 1)]
)
def test_point_location_check(session, bc_geom, nof_invalid):
    factories.ConnectionNodeFactory(id=0, geom=f"SRID={SRID};POINT({POINT1})")
    factories.BoundaryConditions1DFactory(
        id=0, connection_node_id=0, geom=f"SRID={SRID};POINT({bc_geom})"
    )
    errors = PointLocationCheck(
        column=models.BoundaryCondition1D.geom,
        ref_column=models.BoundaryCondition1D.connection_node_id,
        ref_table=models.ConnectionNode,
        max_distance=1,
    ).get_invalid(session)
    assert len(errors) == nof_invalid


@pytest.mark.parametrize(
    "channel_geom, nof_invalid",
    [
        (f"LINESTRING({POINT1}, {POINT3})", 0),
        (f"LINESTRING({POINT2}, {POINT3})", 0),  # within tolerance
        (f"LINESTRING({POINT3}, {POINT1})", 0),  # reversed
        (f"LINESTRING({POINT3}, {POINT2})", 0),  # within tolerance, within tolerance
        (f"LINESTRING(142732 473443, {POINT3})", 1),  # startpoint is wrong
        (f"LINESTRING({POINT1}, 142752 473443)", 1),  # endpoint is wrong
    ],
)
def test_linestring_location_check(session, channel_geom, nof_invalid):
    factories.ConnectionNodeFactory(id=1, geom=f"SRID={SRID};POINT({POINT1})")
    factories.ConnectionNodeFactory(id=2, geom=f"SRID={SRID};POINT({POINT3})")
    factories.ChannelFactory(
        connection_node_id_start=1,
        connection_node_id_end=2,
        geom=f"SRID={SRID};{channel_geom}",
    )
    errors = LinestringLocationCheck(
        column=models.Channel.geom,
        ref_column_start=models.Channel.connection_node_id_start,
        ref_column_end=models.Channel.connection_node_id_end,
        ref_table_start=models.ConnectionNode,
        ref_table_end=models.ConnectionNode,
        max_distance=1.01,
    ).get_invalid(session)
    assert len(errors) == nof_invalid


@pytest.mark.parametrize(
    "channel_geom, nof_invalid",
    [
        (f"LINESTRING({POINT1}, {POINT3})", 0),
        (f"LINESTRING(142732 473443, {POINT3})", 1),  # startpoint is wrong
    ],
)
def test_connection_node_linestring_location_check(session, channel_geom, nof_invalid):
    factories.ConnectionNodeFactory(id=1, geom=f"SRID={SRID};POINT({POINT1})")
    factories.ConnectionNodeFactory(id=2, geom=f"SRID={SRID};POINT({POINT3})")
    factories.ChannelFactory(
        connection_node_id_start=1,
        connection_node_id_end=2,
        geom=f"SRID={SRID};{channel_geom}",
    )
    errors = ConnectionNodeLinestringLocationCheck(
        column=models.Channel.geom, max_distance=1.01
    ).get_invalid(session)
    assert len(errors) == nof_invalid


@pytest.mark.parametrize(
    "channel_geom, nof_invalid",
    [
        (f"LINESTRING({POINT1}, {POINT3})", 0),
        (f"LINESTRING(142732 473443, {POINT3})", 1),  # startpoint is wrong
    ],
)
@pytest.mark.parametrize(
    "control_table, control_type",
    [(models.MemoryControl, "memory"), (models.TableControl, "table")],
)
def test_control_measure_map_linestring_map_location_check(
    session, control_table, control_type, channel_geom, nof_invalid
):
    factories.MeasureLocationFactory(id=1, geom=f"SRID={SRID};POINT({POINT1})")
    factories.MemoryControlFactory(id=1, geom=f"SRID={SRID};POINT({POINT3})")
    factories.TableControlFactory(id=1, geom=f"SRID={SRID};POINT({POINT3})")
    factories.MeasureMapFactory(
        measure_location_id=1,
        control_id=1,
        control_type="memory",
        geom=f"SRID={SRID};{channel_geom}",
    )
    factories.MeasureMapFactory(
        measure_location_id=1,
        control_id=1,
        control_type="table",
        geom=f"SRID={SRID};{channel_geom}",
    )
    errors = MeasureMapLinestringMapLocationCheck(
        control_table=control_table,
        filters=models.MeasureMap.control_type == control_type,
        max_distance=1.01,
    ).get_invalid(session)
    assert len(errors) == nof_invalid


@pytest.mark.parametrize(
    "channel_geom, nof_invalid",
    [
        (f"LINESTRING({POINT1}, {POINT3})", 0),
        (f"LINESTRING(142732 473443, {POINT3})", 1),  # startpoint is wrong
    ],
)
def test_dwf_map_linestring_map_location_check(session, channel_geom, nof_invalid):
    factories.ConnectionNodeFactory(id=1, geom=f"SRID={SRID};POINT({POINT1})")

    factories.DryWeatherFlowFactory(
        id=1,
        geom=f"SRID={SRID};POLYGON(({POINT1}, {POINT3}, 142742 473445, {POINT1}))",
    )
    factories.DryWheatherFlowMapFactory(
        connection_node_id=1, dry_weather_flow_id=1, geom=f"SRID={SRID};{channel_geom}"
    )
    errors = DWFMapLinestringLocationCheck(max_distance=1.01).get_invalid(session)
    assert len(errors) == nof_invalid


@pytest.mark.parametrize(
    "channel_geom, nof_invalid",
    [
        (f"LINESTRING({POINT1}, {POINT3})", 0),
        (f"LINESTRING(142732 473443, {POINT3})", 1),  # startpoint is wrong
    ],
)
def test_pump_map_linestring_map_location_check(session, channel_geom, nof_invalid):
    factories.ConnectionNodeFactory(id=1, geom=f"SRID={SRID};POINT({POINT1})")
    factories.PumpFactory(id=1, geom=f"SRID={SRID};POINT({POINT3})")
    factories.PumpMapFactory(
        connection_node_id_end=1, pump_id=1, geom=f"SRID={SRID};{channel_geom}"
    )
    errors = PumpMapLinestringLocationCheck(max_distance=1.01).get_invalid(session)
    assert len(errors) == nof_invalid


@pytest.mark.parametrize(
    "channel_geom, nof_invalid",
    [
        (f"LINESTRING({POINT1}, {POINT3})", 0),
        (f"LINESTRING(142732 473443, {POINT3})", 1),  # startpoint is wrong
    ],
)
def test_surface_map_linestring_map_location_check(session, channel_geom, nof_invalid):
    factories.ConnectionNodeFactory(id=1, geom=f"SRID={SRID};POINT({POINT1})")
    factories.SurfaceFactory(
        id=1,
        geom=f"SRID={SRID};POLYGON(({POINT1}, {POINT3}, 142742 473445, {POINT1}))",
    )
    factories.SurfaceMapFactory(
        connection_node_id=1, surface_id=1, geom=f"SRID={SRID};{channel_geom}"
    )
    errors = SurfaceMapLinestringLocationCheck(max_distance=1.01).get_invalid(session)
    assert len(errors) == nof_invalid



from abc import ABC, abstractmethod, abstractproperty, abstractstaticmethod
from typing import Optional, Tuple


class RasterInterface(ABC):
    class NotAvailable(Exception):
        pass

    class NoData(Exception):
        pass

    def __init__(self, path):
        self.path = str(path)

    @abstractstaticmethod
    def available():
        pass

    @abstractmethod
    def _open(self):
        pass

    @abstractmethod
    def _close(self):
        pass

    def __enter__(self) -> "RasterInterface":
        self._open()
        return self

    def __exit__(self, *args, **kwargs):
        self._close()

    @abstractproperty
    def is_valid_geotiff(self) -> bool:
        pass

    @abstractproperty
    def band_count(self) -> int:
        pass

    @abstractproperty
    def has_projection(self) -> bool:
        pass

    @abstractproperty
    def is_geographic(self) -> bool:
        pass

    @abstractproperty
    def epsg_code(self) -> Optional[int]:
        pass

    @abstractproperty
    def pixel_size(self) -> Tuple[Optional[float], Optional[float]]:
        gt = self._dataset.GetGeoTransform()
        if gt is None:
            return None, None
        else:
            return abs(gt[1]), abs(gt[5])

    @abstractproperty
    def min_max(self) -> Tuple[Optional[float], Optional[float]]:
        pass

    @abstractproperty
    def shape(self) -> Tuple[int, int]:
        pass

    @abstractproperty
    def compression(self) -> str:
        pass



import pytest
from threedi_schema import constants, models

from threedi_modelchecker.checks.cross_section_definitions import (
    cross_section_configuration_for_record,
    cross_section_configuration_not_tabulated,
    cross_section_configuration_tabulated,
    CrossSectionCSVFormatCheck,
    CrossSectionExpectEmptyCheck,
    CrossSectionFirstElementNonZeroCheck,
    CrossSectionFirstElementZeroCheck,
    CrossSectionFrictionCorrectLengthCheck,
    CrossSectionGreaterZeroCheck,
    CrossSectionIncreasingCheck,
    CrossSectionListCheck,
    CrossSectionMinimumDiameterCheck,
    CrossSectionNullCheck,
    CrossSectionTableCheck,
    CrossSectionTableColumnIdx,
    CrossSectionVariableFrictionRangeCheck,
    CrossSectionVegetationCorrectLengthCheck,
    CrossSectionVegetationTableNotNegativeCheck,
    CrossSectionYZCoordinateCountCheck,
    CrossSectionYZHeightCheck,
    CrossSectionYZIncreasingWidthIfOpenCheck,
    get_widths_heights_for_tabulated_record,
    OpenIncreasingCrossSectionConveyanceFrictionCheck,
    OpenIncreasingCrossSectionVariableCheck,
)

from . import factories


@pytest.mark.parametrize(
    "cross_section_table, col_idx, expected",
    [
        ("0,0\n2,1\n4,2", CrossSectionTableColumnIdx.height, [0.0, 2.0, 4.0]),
        ("0,0\n2,1\n4,2", CrossSectionTableColumnIdx.width, [0.0, 1.0, 2.0]),
        (
            "0,0\n2,1\n4,2",
            CrossSectionTableColumnIdx.all,
            ([0.0, 2.0, 4.0], [0.0, 1.0, 2.0]),
        ),
        (
            "0,1",
            CrossSectionTableColumnIdx.height,
            [
                0,
            ],
        ),
        (
            "0,1",
            CrossSectionTableColumnIdx.width,
            [
                1,
            ],
        ),
        (
            "0,1",
            CrossSectionTableColumnIdx.all,
            (
                [
                    0,
                ],
                [
                    1,
                ],
            ),
        ),
        ("0\n1", CrossSectionTableColumnIdx.height, [0, 1]),
        ("0\n1", CrossSectionTableColumnIdx.width, []),
        ("foo", CrossSectionTableColumnIdx.width, []),
        ("0", CrossSectionTableColumnIdx.height, [0]),
        ("0", CrossSectionTableColumnIdx.width, []),
    ],
)
def test_parse_cross_section_table(session, cross_section_table, col_idx, expected):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
    )
    # can use any class here!
    check = CrossSectionNullCheck(
        column=models.CrossSectionLocation.cross_section_table
    )
    values = list(check.parse_cross_section_table(session=session, col_idx=col_idx))
    if expected:
        assert values[0][1] == expected
    else:
        assert not values


@pytest.mark.parametrize(
    "vegetation_table, expected",
    [
        ("0,0,0,0", [[0.0, 0.0, 0.0, 0.0]]),
        ("0,0", None),
        ("0,0,0,0\n1,2,3,4", [[0.0, 0.0, 0.0, 0.0], [1.0, 2.0, 3.0, 4.0]]),
    ],
)
def test_parse_cross_section_vegetation_table(session, vegetation_table, expected):
    factories.CrossSectionLocationFactory(
        cross_section_vegetation_table=vegetation_table,
    )
    check = CrossSectionNullCheck(
        column=models.CrossSectionLocation.cross_section_vegetation_table
    )
    values = list(check.parse_cross_section_vegetation_table(session=session))
    if expected:
        assert values[0][1] == expected
    else:
        assert not values


def test_filter_shapes(session):
    # should only check records of given types
    factories.CrossSectionLocationFactory(
        cross_section_width=None,
        cross_section_shape=constants.CrossSectionShape.CIRCLE,
    )
    check = CrossSectionNullCheck(
        column=models.CrossSectionLocation.cross_section_width,
        shapes=[constants.CrossSectionShape.RECTANGLE],
    )
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0

    check = CrossSectionNullCheck(
        column=models.CrossSectionLocation.cross_section_width,
        shapes=[constants.CrossSectionShape.CIRCLE],
    )
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("val, is_valid", [(None, False), (1, True)])
def test_check_null_check(session, val, is_valid):
    factories.CrossSectionLocationFactory(cross_section_height=val)
    # Check if the invalid row is identified
    check = CrossSectionNullCheck(
        column=models.CrossSectionLocation.cross_section_height
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == is_valid


@pytest.mark.parametrize("val, is_valid", [(None, True), (1, False)])
def test_check_expect_empty_check(session, val, is_valid):
    factories.CrossSectionLocationFactory(cross_section_height=val)
    check = CrossSectionExpectEmptyCheck(
        column=models.CrossSectionLocation.cross_section_height
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == is_valid


@pytest.mark.parametrize("width, valid", [(-1, False), (0, False), (1, True)])
def test_check_greater_zero(session, width, valid):
    factories.CrossSectionLocationFactory(
        cross_section_width=width,
    )
    check = CrossSectionGreaterZeroCheck(
        column=models.CrossSectionLocation.cross_section_width
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize(
    "cross_section_table",
    ["0 1", "3;5", "foo", "1,2\n3,", ",2", ",2\n3,4"],
)
def test_csv_format_check_invalid(session, cross_section_table):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
    )
    check = CrossSectionCSVFormatCheck(models.CrossSectionLocation.cross_section_table)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("cross_section_table", [None, "", "0,1", "0,1\n0,1"])
def test_csv_format_check_valid(session, cross_section_table):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
    )
    check = CrossSectionCSVFormatCheck(models.CrossSectionLocation.cross_section_table)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize(
    "table, ncol, valid",
    [("1,2\n3,4", 2, True), ("1,2\n3,4", 3, False), ("1,2\n3,", 2, False)],
)
def test_table_check(session, table, ncol, valid):
    factories.CrossSectionLocationFactory(
        cross_section_table=table,
    )
    check = CrossSectionTableCheck(
        column=models.CrossSectionLocation.cross_section_table, ncol=ncol
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize("table, valid", [("1,2\n3,4", False), ("1,2,4,5", True)])
def test_list_check(session, table, valid):
    factories.CrossSectionLocationFactory(
        cross_section_table=table,
    )
    check = CrossSectionListCheck(
        column=models.CrossSectionLocation.cross_section_table
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize(
    "cross_section_table, valid", [("2,1\n1,1", False), ("1,1\n2,1", True)]
)
def test_increasing_elements(session, cross_section_table, valid):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
    )
    check = CrossSectionIncreasingCheck(
        column=models.CrossSectionLocation.cross_section_table
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize(
    "cross_section_table, valid", [("1,2\n1,1", False), ("0,1\n1,2", True)]
)
def test_first_element_zero_check(session, cross_section_table, valid):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
    )
    check = CrossSectionFirstElementZeroCheck(
        column=models.CrossSectionLocation.cross_section_table
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize("cross_section_table, valid", [("1,0", False), ("1,2", True)])
def test_first_nonzero(session, valid, cross_section_table):
    factories.CrossSectionLocationFactory(
        cross_section_shape=constants.CrossSectionShape.TABULATED_RECTANGLE,
        cross_section_table=cross_section_table,
    )
    check = CrossSectionFirstElementNonZeroCheck(
        column=models.CrossSectionLocation.cross_section_table,
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize(
    "cross_section_table, valid",
    [
        ("0,0\n0,1\n,0,2", True),
        ("0,0", True),
        ("0,1\n0,2\n0,3", False),
        ("0,0\n0,-1\n0,1", False),
    ],
)
def test_check_yz_height(session, cross_section_table, valid):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
    )
    check = CrossSectionYZHeightCheck(
        column=models.CrossSectionLocation.cross_section_table
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize(
    "cross_section_table, valid",
    [
        ("0,0.5\n0.5,0", False),
        ("0,0.5\n0.5,0\n0,0.5", False),
        ("0,0.5\n0.5,0\n1,0\n1.5,0.5", True),
        ("0.5,0\n0,1\n0.5,2\n1.5,2\n1.5,0\n0.5,0", True),
    ],
)
def test_check_yz_coord_count(session, cross_section_table, valid):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
    )
    check = CrossSectionYZCoordinateCountCheck(
        column=models.CrossSectionLocation.cross_section_table
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize(
    "cross_section_table, valid",
    [
        ("0,0.5\n0.5,0\n1,0\n1,0.5", False),
        ("0.5,0\n0,1\n0.5,2\n1.5,2\n1.5,0\n0.5,1", False),
        ("0,0.5\n0.5,0\n1,0\n1.5,0.5", True),
        ("0.5,0\n0,1\n0.5,2\n1.5,2\n1.5,0\n0.5,0", True),
    ],
)
def test_check_yz_increasing_if_open(session, cross_section_table, valid):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
    )
    check = CrossSectionYZIncreasingWidthIfOpenCheck(
        column=models.CrossSectionLocation.cross_section_table
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize(
    "shape,width,height,expected_result",
    [
        (
            constants.CrossSectionShape.RECTANGLE,
            "0.1",
            "0.2",
            0,
        ),  # closed rectangle, sufficient width and height, pass
        (
            constants.CrossSectionShape.CLOSED_RECTANGLE,
            "0.05",
            "0.2",
            1,
        ),  # closed rectangle, insufficient width, fail
        (
            constants.CrossSectionShape.CLOSED_RECTANGLE,
            "0.1",
            "0.03",
            1,
        ),  # closed rectangle, insufficient height, fail
        (
            constants.CrossSectionShape.RECTANGLE,
            "0.1",
            None,
            0,
        ),  # open rectangle, sufficient width, no height, pass
        (None, None, None, 0),  # no shape, should skip tests
    ],
)
def test_check_cross_section_minimum_diameter(
    session, shape, width, height, expected_result
):
    factories.CrossSectionLocationFactory(
        cross_section_width=width,
        cross_section_height=height,
        cross_section_shape=shape,
    )
    check = CrossSectionMinimumDiameterCheck(column=models.CrossSectionLocation.id)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == expected_result


@pytest.mark.parametrize(
    "shape,width,height,expected_result",
    [
        (
            constants.CrossSectionShape.CLOSED_RECTANGLE,
            "0.1",
            "0.2",
            1,
        ),  # closed rectangle, fail
        (constants.CrossSectionShape.RECTANGLE, "0.1", None, 0),  # open rectangle, pass
        (None, None, None, 0),
    ],
)
@pytest.mark.parametrize(
    "friction_type,conveyance",
    [
        (constants.FrictionType.CHEZY, False),
        (constants.FrictionType.MANNING, False),
        (constants.FrictionType.CHEZY_CONVEYANCE, True),
        (constants.FrictionType.MANNING_CONVEYANCE, True),
    ],
)
def test_check_cross_section_increasing_open_with_conveyance_friction(
    session, shape, width, height, expected_result, friction_type, conveyance
):
    factories.CrossSectionLocationFactory(
        cross_section_width=width,
        cross_section_height=height,
        cross_section_shape=shape,
        friction_type=friction_type,
    )
    check = OpenIncreasingCrossSectionConveyanceFrictionCheck(
        column=models.CrossSectionLocation.id
    )
    # this check should pass on cross-section locations which don't use conveyance,
    # regardless of their other parameters
    if not conveyance:
        expected_result = 0
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == expected_result


@pytest.mark.parametrize(
    "cross_section_table,expected_result",
    [
        (
            "0.04,0.06\n0.1,0.2",
            0,
        ),  # open tabulated rectangle, increasing width, pass
        (
            "0.04,0.06\n0.1,0.2\n0.1,0.3",
            0,
        ),  # open tabulated rectangle, equal width segments, pass
        (
            "0.2,0.06\n0.1,0.2",
            1,
        ),  # open tabulated rectangle, decreasing width, fail
        (
            "0.04,0.06\n0.1,0.2\n0,0",
            1,
        ),  # closed tabulated rectangle, fail
    ],
)
@pytest.mark.parametrize(
    "friction_type,conveyance",
    [
        (constants.FrictionType.CHEZY, False),
        (constants.FrictionType.MANNING, False),
        (constants.FrictionType.CHEZY_CONVEYANCE, True),
        (constants.FrictionType.MANNING_CONVEYANCE, True),
    ],
)
def test_check_cross_section_increasing_open_with_conveyance_friction_tabulated(
    session, cross_section_table, expected_result, friction_type, conveyance
):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
        cross_section_shape=constants.CrossSectionShape.TABULATED_YZ,
        friction_type=friction_type,
    )
    check = OpenIncreasingCrossSectionConveyanceFrictionCheck(
        column=models.CrossSectionLocation.id
    )
    # this check should pass on cross-section locations which don't use conveyance,
    # regardless of their other parameters
    if not conveyance:
        expected_result = 0
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == expected_result


@pytest.mark.parametrize("data, result", [["1,2", True], ["1,2,3", False]])
def test_check_correct_length_friction(session, data, result):
    factories.CrossSectionLocationFactory(
        cross_section_table="1,0\n2,2\n3,5",
        cross_section_friction_values=data,
    )
    check = CrossSectionFrictionCorrectLengthCheck(
        column=models.CrossSectionLocation.cross_section_friction_values
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == result


@pytest.mark.parametrize("nrows, result", [[2, True], [3, False]])
def test_check_correct_lengthtest_check_vegetation(session, nrows, result):
    factories.CrossSectionLocationFactory(
        cross_section_table="1,0\n2,2\n3,5",
        cross_section_vegetation_table="\n".join(nrows * ["1,2,3,4"]),
    )
    check = CrossSectionVegetationCorrectLengthCheck(
        column=models.CrossSectionLocation.cross_section_vegetation_table
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == result


@pytest.mark.parametrize(
    "table, valid", [("0,0,0,0", True), ("1,2,3,4", True), ("0,0,0,-1", False)]
)
def test_cross_section_vegetation_table_not_negative_check(session, table, valid):
    factories.CrossSectionLocationFactory(cross_section_vegetation_table=table)
    check = CrossSectionVegetationTableNotNegativeCheck(
        column=models.CrossSectionLocation.cross_section_vegetation_table
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == valid


@pytest.mark.parametrize(
    "friction_types, result",
    [
        [[constants.FrictionType.MANNING], False],
        [[constants.FrictionType.CHEZY], True],
    ],
)
def test_check_friction_values_range(session, friction_types, result):
    factories.CrossSectionLocationFactory(
        cross_section_friction_values="0,1",
        friction_type=constants.FrictionType.MANNING,
    )
    check = CrossSectionVariableFrictionRangeCheck(
        min_value=0,
        max_value=1,
        right_inclusive=False,
        error_code=9999,
        column=models.CrossSectionLocation.cross_section_friction_values,
        friction_types=friction_types,
    )
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == result


@pytest.mark.parametrize(
    "cross_section_table,result",
    [
        (
            "0.01,0.11\n0.11,0.21",
            True,
        ),  # open tabulated yz, increasing width, pass
        (
            "0.11,0.11\n0.01,0.20",
            False,
        ),  # open tabulated yz, decreasing width, fail
        (
            "0.01,0.11\n0.11,0.21\n0.01,0.11",
            False,
        ),  # closed tabulated yz,  fail
    ],
)
def test_check_cross_section_increasing_open_with_variables(
    session, cross_section_table, result
):
    factories.CrossSectionLocationFactory(
        cross_section_table=cross_section_table,
        cross_section_friction_values="1",
        cross_section_shape=constants.CrossSectionShape.TABULATED_YZ,
    )
    check = OpenIncreasingCrossSectionVariableCheck(
        models.CrossSectionLocation.cross_section_friction_values
    )
    # this check should pass on cross-section locations which don't use conveyance,
    # regardless of their other parameters
    invalid_rows = check.get_invalid(session)
    assert (len(invalid_rows) == 0) == result


@pytest.mark.parametrize(
    "shape,width,height,expected",
    [
        (constants.CrossSectionShape.CLOSED_RECTANGLE, 1, 2, (1, 2, "closed")),
        (constants.CrossSectionShape.CLOSED_RECTANGLE, None, None, (0, 0, "closed")),
        (constants.CrossSectionShape.RECTANGLE, 1, 2, (1, 2, "open")),
        (constants.CrossSectionShape.RECTANGLE, None, None, (0, None, "open")),
        (constants.CrossSectionShape.CIRCLE, 1, 2, (1, 1, "closed")),
        (constants.CrossSectionShape.CIRCLE, None, None, (0, 0, "closed")),
        (constants.CrossSectionShape.CIRCLE, None, 2, (0, 0, "closed")),
        (constants.CrossSectionShape.INVERTED_EGG, 1, 2, (1, 1.5, "closed")),
        (constants.CrossSectionShape.EGG, None, None, (0, 0, "closed")),
        (constants.CrossSectionShape.EGG, None, 2, (0, 0, "closed")),
    ],
)
def test_cross_section_configuration_not_tabulated(shape, width, height, expected):
    assert cross_section_configuration_not_tabulated(shape, width, height) == expected


@pytest.mark.parametrize(
    "shape,widths,heights,expected",
    [
        (constants.CrossSectionShape.TABULATED_RECTANGLE, None, None, (0, 0, "closed")),
        (
            constants.CrossSectionShape.TABULATED_TRAPEZIUM,
            [1, 2, 3],
            [1, 2, 4],
            (3, 4, "open"),
        ),
        (
            constants.CrossSectionShape.TABULATED_TRAPEZIUM,
            [1, 2, 0],
            [1, 2, 4],
            (2, 4, "closed"),
        ),
        (
            constants.CrossSectionShape.TABULATED_TRAPEZIUM,
            [1, 2, 3],
            [1, 2, 0],
            (3, 2, "open"),
        ),
        (
            constants.CrossSectionShape.TABULATED_YZ,
            [0, 1, 2, 0],
            [1, 2, 3, 4],
            (2, 3, "open"),
        ),
        (
            constants.CrossSectionShape.TABULATED_YZ,
            [0, 1, 2, 0],
            [0, 2, 3, 0],
            (2, 3, "closed"),
        ),
    ],
)
def test_cross_section_configuration_tabulated(shape, widths, heights, expected):
    assert cross_section_configuration_tabulated(shape, widths, heights) == expected


def test_cross_section_configuration_tabulated_raise():
    with pytest.raises(ValueError):
        cross_section_configuration_tabulated(
            shape=constants.CrossSectionShape.RECTANGLE, widths=None, heights=None
        )


def test_cross_section_configuration_not_tabulated_raise():
    with pytest.raises(ValueError):
        cross_section_configuration_not_tabulated(
            shape=constants.CrossSectionShape.TABULATED_YZ, width=None, height=None
        )


@pytest.mark.parametrize(
    "shape,kwargs, expected",
    [
        (
            constants.CrossSectionShape.TABULATED_YZ,
            {"cross_section_table": "0,1\n1,2\n2,3\n0,4"},
            (2, 3, "open"),
        ),
        (
            constants.CrossSectionShape.CLOSED_RECTANGLE,
            {"cross_section_width": 1, "cross_section_height": 2},
            (1, 2, "closed"),
        ),
        (
            constants.CrossSectionShape.TABULATED_YZ,
            {},
            (None, None, None),
        ),
        (
            constants.CrossSectionShape.TABULATED_YZ,
            {"cross_section_table": ""},
            (None, None, None),
        ),
    ],
)
def test_cross_section_configuration_for_record(session, shape, kwargs, expected):
    factories.CrossSectionLocationFactory(cross_section_shape=shape, **kwargs)
    check = CrossSectionNullCheck(models.CrossSectionLocation.id)
    records = list(check.to_check(session))
    assert cross_section_configuration_for_record(records[0]) == expected


@pytest.mark.parametrize(
    "shape, table, expected_widths, expected_heights",
    [
        (
            constants.CrossSectionShape.TABULATED_YZ,
            "0,1\n1,2\n2,3\n0,4",
            [0, 1, 2, 0],
            [1, 2, 3, 4],
        ),
        (
            constants.CrossSectionShape.TABULATED_RECTANGLE,
            "0,1\n1,2\n2,3\n0,4",
            [1, 2, 3, 4],
            [0, 1, 2, 0],
        ),
        (
            constants.CrossSectionShape.TABULATED_TRAPEZIUM,
            "0,1\n1,2\n2,3\n0,4",
            [1, 2, 3, 4],
            [0, 1, 2, 0],
        ),
    ],
)
def test_get_widths_heights_for_tabulated_record(
    session, shape, table, expected_widths, expected_heights
):
    factories.CrossSectionLocationFactory(
        cross_section_shape=shape, cross_section_table=table
    )
    check = CrossSectionNullCheck(models.CrossSectionLocation.id)
    records = list(check.to_check(session))
    widths, heights = get_widths_heights_for_tabulated_record(records[0])
    assert widths == expected_widths
    assert heights == expected_heights


def test_get_widths_heights_for_tabulated_record_raise(session):
    factories.CrossSectionLocationFactory(
        cross_section_shape=constants.CrossSectionShape.RECTANGLE
    )
    check = CrossSectionNullCheck(models.CrossSectionLocation.id)
    records = list(check.to_check(session))
    with pytest.raises(ValueError):
        get_widths_heights_for_tabulated_record(records[0])



from typing import List, NamedTuple

from geoalchemy2.functions import ST_Distance, ST_NPoints, ST_PointN
from sqlalchemy.orm import aliased, Session
from threedi_schema.domain import models

from threedi_modelchecker.checks.base import BaseCheck


class PointLocationCheck(BaseCheck):
    """Check if cross section locations are within {max_distance} of their channel."""

    def __init__(
        self,
        ref_column,
        ref_table,
        max_distance,
        *args,
        **kwargs,
    ):
        self.max_distance = max_distance
        self.ref_column = ref_column
        self.ref_table = ref_table
        super().__init__(*args, **kwargs)

    def get_invalid(self, session):
        # get all channels with more than 1 cross section location
        return (
            self.to_check(session)
            .join(
                self.ref_table,
                self.ref_table.id == self.ref_column,
            )
            .filter(ST_Distance(self.column, self.ref_table.geom) > self.max_distance)
            .all()
        )

    def description(self):
        return (
            f"{self.column_name} does not match the position of the object that "
            f"{self.table.name}.{self.ref_column} refers to"
        )


class LinestringLocationCheck(BaseCheck):
    """Check that linestring geometry starts / ends are close to their connection nodes

    This allows for reversing the geometries. threedi-gridbuilder will reverse the geometries if
    that lowers the distance to the connection nodes.
    """

    def __init__(
        self,
        ref_column_start,
        ref_column_end,
        ref_table_start,
        ref_table_end,
        max_distance,
        *args,
        **kwargs,
    ):
        self.max_distance = max_distance
        self.ref_column_start = ref_column_start
        self.ref_column_end = ref_column_end
        self.ref_table_start = ref_table_start
        self.ref_table_end = ref_table_end
        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        start_node = aliased(self.ref_table_start)
        end_node = aliased(self.ref_table_end)
        tol = self.max_distance
        start_point = ST_PointN(self.column, 1)
        end_point = ST_PointN(self.column, ST_NPoints(self.column))
        start_ok = ST_Distance(start_point, start_node.geom) <= tol
        end_ok = ST_Distance(end_point, end_node.geom) <= tol
        start_ok_if_reversed = ST_Distance(end_point, start_node.geom) <= tol
        end_ok_if_reversed = ST_Distance(start_point, end_node.geom) <= tol
        return (
            self.to_check(session)
            .join(start_node, start_node.id == self.ref_column_start)
            .join(end_node, end_node.id == self.ref_column_end)
            .filter(
                ~(start_ok & end_ok),
                ~(start_ok_if_reversed & end_ok_if_reversed),
            )
            .all()
        )

    def description(self) -> str:
        ref_start_name = f"{self.table.name}.{self.ref_column_start.name}"
        ref_end_name = f"{self.table.name}.{self.ref_column_end.name}"
        return f"{self.column_name} does not start or end at its connection nodes: {ref_start_name} and {ref_end_name} (tolerance = {self.max_distance} m)"


class ConnectionNodeLinestringLocationCheck(LinestringLocationCheck):
    def __init__(self, column, *args, **kwargs):
        table = column.table
        super().__init__(
            ref_column_start=table.c.connection_node_id_start,
            ref_column_end=table.c.connection_node_id_end,
            ref_table_start=models.ConnectionNode,
            ref_table_end=models.ConnectionNode,
            column=column,
            *args,
            **kwargs,
        )

    def description(self) -> str:
        return f"{self.column_name} does not start or end at its connection node (tolerance = {self.max_distance} m)"


class MeasureMapLinestringMapLocationCheck(LinestringLocationCheck):
    def __init__(self, control_table, filters, *args, **kwargs):
        super().__init__(
            ref_column_start=models.MeasureMap.measure_location_id,
            ref_column_end=models.MeasureMap.control_id,
            ref_table_start=models.MeasureLocation,
            ref_table_end=control_table,
            column=models.MeasureMap.geom,
            filters=filters,
            *args,
            **kwargs,
        )


class DWFMapLinestringLocationCheck(LinestringLocationCheck):
    def __init__(self, *args, **kwargs):
        super().__init__(
            ref_column_start=models.DryWeatherFlowMap.connection_node_id,
            ref_column_end=models.DryWeatherFlowMap.dry_weather_flow_id,
            ref_table_start=models.ConnectionNode,
            ref_table_end=models.DryWeatherFlow,
            column=models.DryWeatherFlowMap.geom,
            *args,
            **kwargs,
        )


class PumpMapLinestringLocationCheck(LinestringLocationCheck):
    def __init__(self, *args, **kwargs):
        super().__init__(
            ref_column_start=models.PumpMap.pump_id,
            ref_column_end=models.PumpMap.connection_node_id_end,
            ref_table_start=models.Pump,
            ref_table_end=models.ConnectionNode,
            column=models.PumpMap.geom,
            *args,
            **kwargs,
        )


class SurfaceMapLinestringLocationCheck(LinestringLocationCheck):
    def __init__(self, *args, **kwargs):
        super().__init__(
            ref_column_start=models.SurfaceMap.surface_id,
            ref_column_end=models.SurfaceMap.connection_node_id,
            ref_table_start=models.Surface,
            ref_table_end=models.ConnectionNode,
            column=models.SurfaceMap.geom,
            *args,
            **kwargs,
        )
